@@ -13,6 +13,8 @@ Covering areas:
 import argparse
 import collections
 import cProfile as profile
+import difflib
+import fnmatch
 import functools
 import gettext
 import hashlib
@@ -22,12 +24,15 @@ import logging
 import logging.config
 import multiprocessing
 import os
+import os.path as osp
 from os.path import abspath, basename, dirname, expanduser, exists, isfile, join, splitext
 # from pprint import pprint, pformat
 import platform
 import plistlib
+import pprint as pp
 import pstats
 import shlex
+import shutil
 import subprocess
 import sys
 import tempfile
@@ -331,6 +336,8 @@ def save_json(path, config):
     Unicode as you write, then use json.dump() to write to that file.
     Validate keys to avoid JSON and program out-of-sync.
     """
+    par_dir = osp.split(path)[0]
+    os.makedirs(par_dir, exist_ok=True)
     if is_python3():
         with open(path, 'w', encoding=TXT_CODEC) as f:
             json.dump(config, f, ensure_ascii=False, indent=4)
@@ -920,24 +927,51 @@ def remove_from_os_paths(bindir):
     os.environ['PATH'] = os.environ['PATH'].replace(bindir, '')
 
 
-def run_cmd(cmd, cwd='.'):
-    try:
-        proc = subprocess.run(cmd, check=True, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
-        _logger.debug(proc.stdout.decode(TXT_CODEC))
-    except subprocess.CalledProcessError as e:
-        _logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
-        raise e
-
-
-def run_daemon(cmd, cwd='.'):
+def run_cmd(cmd, cwd='.', logger=None):
+    if logger:
+            logger.debug(' '.join(cmd))
+    else:
+        print(' '.join(cmd))
     proc = None
     try:
-        proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
+        proc = subprocess.run(cmd, check=True, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
+        log = proc.stdout.decode(TXT_CODEC)
+        if logger:
+            logger.debug(log)
+        else:
+            print(log)
     except subprocess.CalledProcessError as e:
-        _logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
+        log = f'stdout: {e.stdout.decode(TXT_CODEC)}'
+        err = f'stderr: {e.stderr.decode(TXT_CODEC)}'
+        if logger:
+            logger.info(log)
+            logger.error(err)
+        else:
+            print(log)
+            print(err)
+        raise e
+    except Exception as e:
+        if logger:
+            logger.error(e)
         raise e
     return proc
 
+def run_daemon(cmd, cwd='.', logger=None):
+    proc = None
+    try:
+        proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
+        if logger:
+            logger.debug(proc.stdout.decode(TXT_CODEC))
+    except subprocess.CalledProcessError as e:
+        if logger:
+            logger.info(f'stdout: {e.stdout.decode(TXT_CODEC)}')
+            logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
+        raise e
+    except Exception as e:
+        if logger:
+            logger.error(e)
+        raise e
+    return proc
 
 
 def extract_call_args(file, caller, callee):
@@ -948,9 +982,11 @@ def extract_call_args(file, caller, callee):
     import ast
     import importlib
     import inspect
-    mod = splitext(basename(file))[0]
+    mod_name = splitext(basename(file))[0]
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
     sys.path.insert(0, dirname(file))
-    mod = importlib.import_module(mod)
+    mod = importlib.import_module(mod_name)
     parsed = ast.parse(inspect.getsource(mod))
     raw_calls = { # lineno, args, keywords
         'func': [],
@@ -960,14 +996,13 @@ def extract_call_args(file, caller, callee):
         if not isinstance(node, ast.FunctionDef) \
             or node.name != caller:
             continue
-        # hit function definition: add_arguments
+        # hit function definition
         # next: find function calls in this function
         caller_def = node
         callee_calls = []
         for node in caller_def.body:
-            if not isinstance(node.value, ast.Call):
+            if 'value' not in dir(node) or not isinstance(node.value, ast.Call):
                 continue
-            print(node.value.func)
             # hit a function/method call
             call_type = None
             found_func_call = isinstance(node.value.func, ast.Name) and node.value.func.id == callee
@@ -1019,12 +1054,15 @@ def extract_call_args(file, caller, callee):
                     print(f'Unsupported syntax node: {kw.value}. Will fallback to None.')
                     value = None
                 kwargs.append((key, value))
-            call = {
+            record = {
                 'args': args,
-                'kwargs': {k: v for k, v in kwargs}
+                'kwargs': {k: v for k, v in kwargs},
+                'lineno': call.lineno,
+                'end_lineno': call.end_lineno,
             }
-            calls[calltype].append(call)
-    sys.modules.pop(mod)
+            calls[calltype].append(record)
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
     return calls['func'], calls['method']
 
 
@@ -1032,15 +1070,19 @@ def extract_class_attributes(file, classname):
     """
     assume
     - class is defined at the top-level of source file
-    - attributes all use type-annotated assignemts
-    - all assgiments are from attributes
+    - all attributes are defined in constructor
+    - all assignments must be about attributes, no local variable are allowed
+    - attributes can use type-annotated assignemts (taa)
+    - types of attributes without taa can be inferred from constant values
     """
     import ast
     import importlib
     import inspect
-    mod = splitext(basename(file))[0]
+    mod_name = splitext(basename(file))[0]
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
     sys.path.insert(0, dirname(file))
-    mod = importlib.import_module(mod)
+    mod = importlib.import_module(mod_name)
     parsed = ast.parse(inspect.getsource(mod))
     
     class_node = None
@@ -1051,57 +1093,357 @@ def extract_class_attributes(file, classname):
             continue
         class_node = node
     if not class_node:
-        return None
-    # found class and parse
-    names = [node.attr for node in ast.walk(class_node) if isinstance(node, ast.Attribute)]
-    types = [node.annotation.id for node in ast.walk(class_node) if isinstance(node, ast.AnnAssign)]
-    values = []
+        return None    
+    ctor = None
     for node in ast.walk(class_node):
-        if not isinstance(node, ast.AnnAssign):
+        if isinstance(node, ast.FunctionDef) and node.name == '__init__':
+            ctor = node
+            break;
+    if not ctor:
+        return None
+    # parse ctor
+    names = [node.attr for node in ast.walk(ctor) if isinstance(node, ast.Attribute)]
+    # types = [node.annotation.id for node in ast.walk(ctor) if isinstance(node, ast.AnnAssign)]
+    types = []
+    values = []
+    linenos = []
+    end_linenos = []
+    for node in ast.walk(ctor):
+        if not isinstance(node, (ast.AnnAssign, ast.Assign)):
             continue
+        attr_type = None
+        attr_value = None
+        if isinstance(node, ast.AnnAssign):
+            attr_type = node.annotation.id
         if isinstance(node.value, ast.Constant):
-            values.append(node.value.value)
+            if not attr_type:
+                attr_type = type(node.value.value).__name__
+            attr_value = node.value.value
         elif isinstance(node.value, (ast.List, ast.Tuple)):
             nodes = node.value.elts
             raw_values = []
-            for node in nodes:
+            for nd in nodes:
                 # only support constants inside list
                 # non-consts are taken as None
-                rv = node.value if isinstance(node, ast.Constant) else None
+                rv = nd.value if isinstance(nd, ast.Constant) else None
                 raw_values.append(rv)
-            values.append(raw_values)
-    attributes = [{'name': n, 'type': t, 'value': v} for n, t, v in zip(names, types, values)]
-    sys.modules.pop(mod)
+            if not attr_type:
+                attr_type = 'list' if isinstance(nd.value, ast.List) else 'tuple'
+            attr_value = raw_values
+        types.append(attr_type)
+        values.append(attr_value)
+        linenos.append(node.lineno)
+        end_linenos.append(node.end_lineno)
+    attributes = [{'name': n, 'type': t, 'default': v, 'lineno': l, 'end_lineno': e} for n, t, v, l, e in zip(names, types, values, linenos, end_linenos)]
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
     return attributes
 
 
-def substitute_lines_between_keywords(lines, file, opkey, edkey):
+def extract_local_var_assignments(file, caller, varname):
     """
-    assume input lines all have line ends
+    - only support regular assignments (var_name = literal_value)
+    """
+    value = None
+    import ast
+    import importlib
+    import inspect
+    mod_name = splitext(basename(file))[0]
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
+    sys.path.insert(0, dirname(file))
+    mod = importlib.import_module(mod_name)
+    parsed = ast.parse(inspect.getsource(mod))
+    raw_calls = { # lineno, args, keywords
+        'func': [],
+        'method': []
+    }
+    assignments = []
+    for node in parsed.body:
+        if not isinstance(node, ast.FunctionDef) \
+            or node.name != caller:
+            continue
+        # hit function definition
+        # next: find function calls in this function
+        func_def = node
+        for node in ast.walk(func_def):
+            if not isinstance(node, ast.Assign) or len(node.targets) > 1:
+                continue
+            var = node.targets[0]
+            if not isinstance(var, ast.Name) or var.id != varname:
+                continue
+            if not isinstance(node.value, ast.Constant):
+                continue
+            ass = {
+                'lineno': node.lineno,
+                'end_lineno': node.end_lineno,
+                'value': node.value.value
+            }
+            assignments.append(ass)
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
+    return assignments
+
+
+def extract_imported_modules(file):
+    def _extract_import_module(node):
+        return [alias.name for alias in node.names]
+
+    def _extract_from_module_import(node):
+        return node.module
+    imported = []
+    import ast
+    import importlib
+    import inspect
+    mod_name = splitext(basename(file))[0]
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
+    sys.path.insert(0, dirname(file))
+    mod = importlib.import_module(mod_name)
+    parsed = ast.parse(inspect.getsource(mod))
+    for node in ast.walk(parsed):
+        if not isinstance(node, (ast.Import, ast.ImportFrom)):
+            continue
+        if isinstance(node, ast.Import):
+            extracted = _extract_import_module(node)
+            if extracted:
+                imported += extracted
+            continue
+        extracted = _extract_from_module_import(node)
+        if extracted:
+            imported.append(extracted)
+    if mod_name in sys.modules:
+        sys.modules.pop(mod_name)
+    return sorted(list(set(imported)))
+
+
+def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0, withindent=True, useappend=False):
+    """
+    - assume input lines all have line ends
+    - align inserted text with tags via identical indents
+    - optimize with search range slicing
+    - returns original indices
     """
     all_lines = []
     with open(file) as fp:
         all_lines = fp.readlines()
+    selected_lines = all_lines[startlineno:] if startlineno > 0 else all_lines
     # find range
-    rg_insert = [-1, -1]
-    for l, line in enumerate(all_lines):
-        raw = line.strip()
-        if raw.startswith(opkey):
-            rg_insert[0] = l
-        if raw.startswith(edkey):
-            rg_insert[1] = l
-            break
-    # remove lines in b/w
-    n_lines_btw = rg_insert[1] - rg_insert[0]
-    if n_lines_btw > 1:
+    rg_insert = [None, None]
+    rg_insert[0] = next((l for l, line in enumerate(selected_lines) if line.strip().startswith(opkey) ), None)
+    if rg_insert[0] is None:
+        return rg_insert
+    rg_insert[1] = next((l for l, line in enumerate(selected_lines[rg_insert[0]:]) if line.strip().startswith(edkey) ), None)
+    if rg_insert[1] is None:
+        return (startlineno+rg_insert[0], None)
+    # back to all lines with offset applied
+    rg_insert[0] += startlineno
+    rg_insert[1] += rg_insert[0]
+    if withindent:
+        indent = all_lines[rg_insert[0]].find(opkey)
+        indent_by_spaces = 0
+        for idt in range(indent):
+            indent_by_spaces += 4 if all_lines[rg_insert[0]][idt] == '\t' else 1
+        assert indent_by_spaces >= 0
+        lines = ['{}{}'.format(' '*indent_by_spaces, line) for line in lines]
+    if useappend:
+        rg_insert[0] = rg_insert[1]-1
+    else:
+        # remove lines in b/w
+        has_lines_between_keywords = rg_insert[1] - rg_insert[0] > 1
         del all_lines[rg_insert[0]+1 : rg_insert[1]]
     all_lines[rg_insert[0]+1 : rg_insert[0]+1] = lines
     with open(file, 'w') as fp:
         fp.writelines(all_lines)
+    rg_inserted = [rg_insert[0], rg_insert[0]+len(lines)]
+    return rg_inserted
+
+
+def convert_compound_cases(snake_text, style='pascal'):
+    if style == 'oneword':
+        return snake_text.replace('_', '').lower()
+    if style == 'ONEWORD':
+        return snake_text.replace('_', '').upper()
+    if style == 'SNAKE':
+        return snake_text.upper()
+    if style == 'kebab':
+        return snake_text.replace('_', '-')
+    out_text = [s.capitalize() for s in snake_text.split('_')]
+    if style == 'camel':
+        out_text[0] = out_text[0].lower()
+    return ''.join(out_text)
+
+
+def combine_words(words, style='snake'):
+    out_text = '_'.join(words)
+    if style == 'snake':
+        return out_text
+    if style == 'sentence':
+        return ' '.join(words)
+    if style == 'Sentence':
+        return ' '.join(words).capitalize()
+    if style == 'title':
+        return ' '.join([word.capitalize() for word in words])
+    return convert_compound_cases(out_text, style)
+
+
+def append_lineends_to_lines(lines, style='posix'):
+    lineend = '\r\n' if style == 'windows' else '\n'
+    return [line + lineend for line in lines]
+
+
+def zip_dir(srcdir, dstbasename=None):
+    """
+    zip the entire folder into a zip file under the same parent folder.
+    """
+    def _zip_dir_windows(srcdir, dstbasename):
+        src_par, src_name = osp.split(srcdir)
+        if not dstbasename:
+            dstbasename = src_name
+        elif dstbasename != src_name:
+            dst_dir = osp.join(src_par, dstbasename)
+            duplicate_dir(srcdir, dst_dir)
+        out_file = osp.join(src_par, dstbasename+'.zip')
+        cmd = ['tar', '-cf', out_file, '-C', src_par, dstbasename]
+        run_cmd(cmd, src_par)
+
+    def _zip_dir_macos(srcdir, dstbasename):
+        src_par, src_name = osp.split(srcdir)
+        rename_option = []
+        if not dstbasename:
+            dstbasename = src_name
+        elif dstbasename != src_name:
+            dst_dir = osp.join(src_par, dstbasename)
+            rename_option = ['-s', f'/^{src_name}/{dstbasename}/']
+        out_file = osp.join(src_par, dstbasename+'.zip')
+        cmd = ['tar', '-czf', out_file, '--exclude', '.DS_Store', '--exclude', '*/__MACOSX'] + rename_option + ['-C', src_par, f'{src_name.strip(os.path.sep)}/']
+        run_cmd(cmd, src_par)
+
+    if platform.system() == 'Windows':
+        _zip_dir_windows(srcdir, dstbasename)
+        return
+    _zip_dir_macos(srcdir, dstbasename)
+
+
+def unzip_dir(srcball, destpardir):
+    """
+    assume srcball has a top-level folder "product".
+    unzip to destpardir/product.
+    """
+    untar_option = '-xf' if platform.system() == 'Windows' else '-xzf'
+    os.makedirs(destpardir, exist_ok=True)
+    cmd = ['tar', untar_option, srcball, '-C', destpardir]
+    run_cmd(cmd, destpardir)
+    
+
+def duplicate_dir(srcdir, dstdir):
+    def _dup_dir_posix(srcdir, dstdir):
+        if not srcdir.endswith('/'):
+            srcdir += '/'
+        cmd = ['rsync', '-a', '--delete', srcdir, dstdir.strip('/')]
+        run_cmd(cmd, '/')
+
+    def _dup_dir_windows(srcdir, dstdir):
+        cmd = ['xcopy', '/I', '/E', '/Q', '/Y', srcdir, f'{dstdir}\\']
+        run_cmd(cmd, srcdir)
+
+    if platform.system() == 'Windows':
+        _dup_dir_windows(srcdir, dstdir)
+        return
+    _dup_dir_posix(srcdir, dstdir)
+
+
+def compare_textfiles(file1, file2, showdiff=False, contextonly=True, logger=None):
+    with open(file1) as fp1, open(file2) as fp2:
+        lines1 = fp1.readlines()
+        lines2 = fp2.readlines()
+        if showdiff:
+            diff_func = difflib.context_diff if contextonly else difflib.Differ().compare
+            diff = diff_func(lines1, lines2)
+            lazy_logging(''.join(diff))
+    assert lines1 == lines2, f'content different: {file1} vs. {file2}'
+
+
+def lazy_logging(msg, logger=None):
+    if logger:
+        logger.info(msg)
+    else:
+        print(msg)
+
+
+def copy_file(src, dst, isdstdir=False):
+    if isdstdir:
+        par_dir = dst
+    else:
+        par_dir = osp.split(dst)[0]
+    os.makedirs(par_dir, exist_ok=True)
+    try:
+        shutil.copy(src, dst)
+    except shutil.SameFileError:
+        _logger.warning(f'source and destination are identical. will not copy: {osp.abspath(src)} -> {osp.abspath(dst)}. skipped.')
+
+
+def compare_dirs(dir1, dir2, ignoreddirpatterns=(), ignoredfilepatterns=(), showdiff=True):
+    def _collect_folders_files(my_dir):
+        n_truncates = len(my_dir)
+        my_dir_contents = {
+            'dirs': [],
+            'files': [],
+        }
+        for root, folders, files in os.walk(my_dir):
+            for folder in folders:
+                folder_matching_pattern = next((pat for pat in ignoreddirpatterns if pat in folder), None)
+                if folder_matching_pattern:
+                    continue
+                my_dir_contents['dirs'].append(osp.join(root, folder)[n_truncates+1:])
+            for file in files:
+                file_matching_pattern = next((pat for pat in ignoredfilepatterns if fnmatch.fnmatch(file, pat)), None)
+                if file_matching_pattern:
+                    continue
+                my_dir_contents['files'].append(osp.join(root, file)[n_truncates+1:])
+        
+        my_dir_contents['dirs'] = sorted(my_dir_contents['dirs'])
+        my_dir_contents['files'] = sorted(my_dir_contents['files'])
+        return my_dir_contents
+    dir1_contents = _collect_folders_files(dir1)
+    dir2_contents = _collect_folders_files(dir2)
+    if showdiff:
+        import filecmp
+        dc = filecmp.dircmp(dir1, dir2, ignore=list(ignoreddirpatterns))
+        dc.report_full_closure()
+        pp.pprint(f'dir1: {dir1_contents}')
+        pp.pprint(f'dir2: {dir2_contents}')
+    assert dir1_contents['dirs'] == dir2_contents['dirs'], 'folders different:\n{}\n\nvs.\n\n{}'.format(pp.pformat(dir1_contents['dirs'], indent=2), pp.pformat(dir2_contents['dirs'], indent=2))
+    assert dir1_contents['files'] == dir2_contents['files'], 'files different:\n{}\n\nvs.\n\n{}'.format(pp.pformat(dir1_contents['files'], indent=2), pp.pformat(dir2_contents['files'], indent=2))
+
+
+def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=()):
+    """
+    for cross-language rpc only, so no need for an unpack()
+    pickle is already good enough for local transmission.
+    """
+    if not topic:
+        topic = type(obj).__name__
+    from types import SimpleNamespace
+    msg = SimpleNamespace(payload=obj, topic=topic)
+    if classes:
+        class CustomJsonEncoder(json.JSONEncoder):
+            def default(self, o):
+                if isinstance(o, classes):
+                    return o.__dict__
+                else:
+                    return json.JSONEncoder.encode(self, o)
+        classes = tuple([SimpleNamespace]+list(classes))
+        msg_str = json.dumps(msg, cls=CustomJsonEncoder)
+    else:
+        msg_str = json.dumps(msg, default=lambda o: o.__dict__)
+
+    return f'{envelope[0]}{msg_str}{envelope[1]}'
 
 
 def _test():
-    print(extract_class_attributes('/Users/bin.luo/Desktop/_dev/miatech/create_py_proj/test/proto_gen/_ref_org/core.py', 'Output'))
+    ret = substitute_lines_between_keywords(['line 1\n', 'line 2\n'], '/Users/bin.luo/Desktop/tmp.txt', '<TAG', 'TAG>', useappend=True)
+    print(ret)
 
 
 if __name__ == '__main__':
